/* written by John Doug Reynolds, January 2002 */

#include "rgsbadpix.h"
#include <PixelsTable.h>
#include <EventAttributes.h>

namespace {
   using namespace rgsPixelsLib;

   class BadPix {
   public:
      enum Flag { Uplinked = 1, Advisory = 2, Found = 3 };

      short chipX, chipY;
      unsigned int yExtent;
      unsigned char status;

      BadPix( short x, short y, unsigned int n, unsigned char tag )
	 : chipX(x), chipY(y), yExtent(n), status(tag)
	 {}

      enum BadPixel::DefectTypes type() const { return type(status); }
      enum Flag flag() const { return flag(status); }

      static unsigned char tag(enum BadPixel::DefectTypes type, enum Flag flag)
	 { return (type << 2) + flag; }

      static enum BadPixel::DefectTypes type( unsigned char tag )
	 { return BadPixel::DefectTypes(tag >> 2); }

      static enum Flag flag( unsigned char tag )
	 { return Flag(tag & 3); }
   };

   class BadPixBuilder : private Contiguator<unsigned char> {
   public:
      BadPixBuilder( vector<BadPix> badpix[2], const PixelsTable&,
		     const Table *frames, struct GeometryDataServer*,
		     int pixnoiselimit, int colnoiselimit,
		     float pixsharpness, float colsharpness,
		     bool withadvisory, bool withfoundhot
		     );

   private:
      struct CalBadPix {
	 long		pixid;
	 unsigned int	extent;
	 unsigned char	tag;
	 CalBadPix( long pixid, unsigned extent, const BadPixel& );
      };

      struct ColumnStats {
	 unsigned	pixels;
	 float		average;
	 ColumnStats() : pixels(0) {}
      };

      const unsigned char Invalid;	// 255
      const unsigned char FoundHot;	// tag(HOT,Found)
      const unsigned char AdvisedHot;	// tag(HOT,Advisory)

      const unsigned	_ccd;
      struct Bounds	_window[2];

      vector<CalBadPix> _calbp;

      struct PixelMap<unsigned int> _pixEventMap;
      struct PixelMap<unsigned int> _colEventMap;

      vector<ColumnStats> _columnStats;

      void initCalBadPix( struct GeometryDataServer*, bool withadvisory );
      unsigned loadEvents( const PixelsTable &ptab, const Table *frames,
			   int pixnoiselimit, int colnoiselimit );
      void markInactives();
      void findHotPixels( unsigned frameCount,
			  float pixsharpness, float colsharpness );
      void markCalBadPix( bool withfoundhot );
      void collect( vector<BadPix> badpix[2] );

      bool badPixel( unsigned counts, unsigned neighborCounts,
		     float sharpness2PerFrame, unsigned frameCount ) const;
      bool badColumn( const ColumnStats&, const ColumnStats &neighbor,
		      float sharpness2PerFrame, unsigned frameCount ) const;
   };

}

void rgsbadpix()
{
   string setname	= stringParameter("ccdset");
   string badpixsetname	= stringParameter("badpixset");
   bool   withbadpixset	= booleanParameter("withbadpixset");
   bool   withfoundhot  = booleanParameter("withfoundhot");

   enum Dal::AccessMode access = withbadpixset ? Dal::Read : Dal::Modify;
   struct DataSet* dataset = dataSetServer->open(setname,access);
   const Table *pixels = dataset->table("PIXELS");

   enum Xmm::DataMode datamode
      = Xmm::toDataMode( dataset->attribute("DATAMODE")->asString() );
   bool spectroscopy = (datamode == Xmm::DATA_SPECTROSCOPY);

   if ( withfoundhot && !spectroscopy ) {
      errstr << "parameter --withfoundhot=true not yet implemented for HTR"
	     << Sas::warning("disabledParameter");
      withfoundhot = false;
   }

   vector<BadPix> badpix[2];
   struct Bounds  bounds[2];
   {
      calServer << *dataset;
      struct GeometryDataServer* calGDS = calServer.getAtom(calGDS);

      unsigned flags = withfoundhot ? PixelsTable::PIX | PixelsTable::PHA : 0;
      struct PixelsTable ptab( pixels, calGDS, flags );

      BadPixBuilder( badpix, ptab, dataset->table("EXPOSURE"), calGDS,
		     intParameter("pixnoiselimit"),
		     intParameter("colnoiselimit"),
		     realParameter("pixsharpness"),
		     realParameter("colsharpness"),
		     booleanParameter("withadvisory"),
		     withfoundhot
		     );

      for ( unsigned n = 0; n < 2; ++n ) bounds[n] = ptab.bounds[n];

      calServer.releaseAtom( calGDS );
   }

   bool abutted = (1 == bounds[1].col.min - bounds[0].col.max);
   struct DataSet *bpset = dataset;

   if ( withbadpixset ) {
      bpset = dataSetServer->open(badpixsetname,Dal::Create);
      bpset->addAttribute( dataset->attribute("TELESCOP") );
      bpset->addAttribute( dataset->attribute("INSTRUME") );
      bpset->addAttribute( dataset->attribute("OBS_ID") );
      bpset->addAttribute( dataset->attribute("EXP_ID") );
   }

   for ( unsigned node = 0; node < 2; ++node ) {
      string extname = node ? "BADPIX1" : "BADPIX0";
      if ( bpset->hasBlock(extname) ) bpset->deleteBlock(extname);
      struct Table *table = bpset->addTable( extname, badpix[node].size() );

      table->addAttribute( dataset->attribute("INSTRUME") );
      table->addAttribute( pixels->attribute("CCDOCB") );
      table->addAttribute( pixels->attribute("CCDID") );

#define ATTR(name,label) *table->addAttribute(name,label)
      ATTR( "CCDNODE","readout node identifier")   = (int) node;
      ATTR("WINDOWX0","node readout origin, CHIPX")= bounds[node].col.min;
      ATTR("WINDOWY0","node readout origin, CHIPY")= bounds[node].row.min;
      ATTR("WINDOWDX","node readout columns")= (int) bounds[node].col.length();
      ATTR("WINDOWDY","node readout rows")   = (int) bounds[node].row.length();
      ATTR( "ABUTTED","abuts on paired node")= abutted;
#undef ATTR

      DalVecI chipY, yExtent;
      DalVecI chipX( table,      "CHIPX", "chip x coordinate", "PIXELS" );
      if ( spectroscopy ) {
	 chipY.reset( table,     "CHIPY", "chip y coordinate", "PIXELS" );
	 yExtent.reset( table, "YEXTENT", "extent of badness in y", "PIXELS" );
      }
      DalVecI  type( table,       "TYPE", "type of badness" );
      DalVecI  flag( table,    "BADFLAG", "data source flag" );

      for ( unsigned i = 0; i < badpix[node].size(); ++i ) {
	 const BadPix &bp = badpix[node][i];
	 chipX[i] = bp.chipX;
	 if ( spectroscopy ) {
	    chipY[i] = bp.chipY;
	    yExtent[i] = (short) bp.yExtent;
	 }
	 type[i] = (short) bp.type();
	 flag[i] = (short) bp.flag();
      }
   }

   if ( withbadpixset ) dataSetServer->close( bpset );
   dataSetServer->close( dataset );
}

/* --- BadPixBuilder methods ----------------------------------------------- */

using namespace Compass;

BadPixBuilder::CalBadPix::CalBadPix( long _pixid, unsigned _extent,
				     const BadPixel &bp )
   : pixid(_pixid), extent(_extent),
     tag( BadPix::tag(bp.type(),
		      bp.uplinked() ? BadPix::Uplinked : BadPix::Advisory) )
{}

BadPixBuilder::BadPixBuilder( vector<BadPix> badpix[2],
			      const PixelsTable &ptab, const Table *frames,
			      struct GeometryDataServer *calGDS,
			      int pixnoiselimit, int colnoiselimit,
			      float pixsharpness, float colsharpness,
			      bool withadvisory, bool withfoundhot
			      )
   : Invalid(255),
     FoundHot(BadPix::tag(BadPixel::HOT,BadPix::Found)),
     AdvisedHot(BadPix::tag(BadPixel::HOT,BadPix::Advisory)),
     _ccd( calServer.state()->ccd()->get() )
{
   for ( unsigned n = 0; n < 2; ++n ) _window[n] = ptab.bounds[n];
   struct Span chipColSpan(_window[0].col.min,_window[1].col.max);
   resize( Bounds(chipColSpan,_window[0].row), 1 );

   initCalBadPix(calGDS,withadvisory);

   if ( withfoundhot ) {
      _pixEventMap.resize( bounds(), 1 );
      _colEventMap.resize( bounds(), 1 );
      _columnStats.resize( 2 + bounds().col.length() );

      unsigned frameCount
	 = loadEvents( ptab, frames, pixnoiselimit, colnoiselimit );

      markInactives();
      findHotPixels( frameCount, pixsharpness, colsharpness );
   }

   markCalBadPix(withfoundhot);
   collect(badpix);
}

void BadPixBuilder::initCalBadPix( struct GeometryDataServer *calGDS,
				   bool withadvisory
				   )
{
   for ( unsigned node = 0; node < 2; ++node ) {
      calServer.state()->node()->set( Node(node) );
      struct BadPixelDataServer *calBPDS = calServer.getAtom(calBPDS);
      const BadPixelMap badpixMap = calBPDS->map();

      // convert node coordinates to chip coordinates

      vector<short> x, y;
      BadPixelMap::const_iterator itr;
      for ( itr = badpixMap.begin(); itr != badpixMap.end(); ++itr ) {
	 const PixelLocation &coordPair = itr->second.where();
	 x.push_back( coordPair.first );
	 y.push_back( coordPair.second );
      }

      CalInt16Vector xV(x.begin(),x.size()), yV(y.begin(),y.size());
      calGDS->pixCoord1List(xV,yV).toChipCoord();

      // clip bad segments at the readout window boundary

      unsigned i = 0;
      for ( itr = badpixMap.begin(); itr != badpixMap.end(); ++itr, ++i ) {
	 if (!_window[0].col.contains(x[i]) && !_window[1].col.contains(x[i]))
	    continue;

	 const BadPixel &bp = itr->second;
	 if ( !withadvisory && !bp.uplinked() ) continue;

	 struct Span rowSpan = Span(y[i]).length(bp.yextent());
	 if ( !(rowSpan &= bounds().row) ) continue;

	 long pixid = id( x[i], rowSpan.min );
	 _calbp.push_back( CalBadPix(pixid,rowSpan.length(),bp) );
      }

      calServer.releaseAtom( calBPDS );
   }
}

unsigned BadPixBuilder::loadEvents( const PixelsTable &ptab, const Table *ftab,
				    int pixnoiselimit, int colnoiselimit )
{
   // initialize badFrameMask and grade-dependent noise thresholds

   int badFrameMask = EventAttributes::CLEAR_ALL_FLAGS;
   EventAttributes::setEventAttribute(badFrameMask,
				      EventAttributes::IN_BAD_FRAME);

   short pixNoiseLimit[4], colNoiseLimit[4];
   struct EnergyCorrector* corrector = calServer.getAtom(corrector);
   vector<float> qSplitR( corrector->qSplitRatios() );
   unsigned i;
   for ( i = 0; i < 4; ++i ) {
      pixNoiseLimit[i] = short(pixnoiselimit / qSplitR[i]);
      colNoiseLimit[i] = short(colnoiselimit / qSplitR[i]);
   }
   calServer.releaseAtom(corrector);

   // simultaneously traverse FRAMES and PIXELS tables

   const DalVecJ frame( ftab->column("FRAME") );
   const DalVecJ flag( ftab->column("FLAG") );

   unsigned frameCount = 0, fItr = 0, pItr = 0;
   while ( fItr < ftab->rows() && pItr < ptab.rows ) {
      int thisFrame = ptab.frame[pItr];

      if ( frame[fItr] < thisFrame ) {
	 if ( !(flag[fItr++] & badFrameMask) ) ++frameCount;
	 continue;
      }

      if ( frame[fItr] != thisFrame || (flag[fItr] & badFrameMask) ) {
	 unsigned start = pItr++;
	 while ( pItr < ptab.rows && ptab.frame[pItr] == thisFrame ) ++pItr;
	 if ( frame[fItr] != thisFrame )
	    errstr << "(CCD " << _ccd << " frame " << thisFrame
		   << ") discarding pixels " << start+1 << " to " << pItr
		   << Sas::warning("missingFrames");
	 else ++fItr;
	 continue;
      }

      // validate each shape and tally its pixels

      for (; pItr < ptab.rows && ptab.frame[pItr] == thisFrame; ++pItr ) {
	 vector<Pixel> pixels;

	 if ( !ptab.valid(pItr,pixels) ) {
	    errstr << "(CCD " << _ccd << " node " << (int)ptab.node[pItr]
		   << ") discarding pixel " << pItr+1 << " at ("
		   << ptab.chipX[pItr] << "," << ptab.chipY[pItr]
		   << ") with shape code " << (int)ptab.shape[pItr]
		   << Sas::warning("invalidCoordinates");
	    continue;
	 }

	 vector<long> ids;
	 vector<long>::iterator id;
	 pixToId( pixels, ids );

	 unsigned g = ptab.grade[pItr] - 1;
	 if ( ptab.energy[pItr] >= pixNoiseLimit[g] )
	    for ( id = ids.begin(); id != ids.end(); ++_pixEventMap[*id++] );
	 if ( ptab.energy[pItr] >= colNoiseLimit[g] )
	    for ( id = ids.begin(); id != ids.end(); ++_colEventMap[*id++] );
      }

      ++frameCount;
      ++fItr;
   }

   if ( pItr < ptab.rows )
      errstr << "(CCD " << _ccd << ") discarding pixels " << pItr+1
	     << " to " << ptab.rows
	     << Sas::warning("missingFrames");

   return frameCount;
}

void BadPixBuilder::markInactives()
{
   struct Bounds b = bounds();
   b.extend( Pixel(b.col.min-1,b.row.min-1) );
   b.extend( Pixel(b.col.max+1,b.row.max+1) );

   PixelMap<unsigned char>::put( Bounds(b.col,Span(b.row.min)), Invalid );
   PixelMap<unsigned char>::put( Bounds(b.col,Span(b.row.max)), Invalid );
   PixelMap<unsigned char>::put( Bounds(Span(b.col.min),b.row), Invalid );
   PixelMap<unsigned char>::put( Bounds(Span(b.col.max),b.row), Invalid );

   struct Span gap(_window[0].col.max+1,_window[1].col.min-1);
   PixelMap<unsigned char>::put( Bounds(gap,b.row), Invalid );

   vector<CalBadPix>::const_iterator itr;
   for ( itr = _calbp.begin(); itr != _calbp.end(); ++itr ) {
      bool uplinked = BadPix::flag(itr->tag) == BadPix::Uplinked;
      if ( !uplinked && BadPix::type(itr->tag) != BadPixel::DEAD ) continue;

      unsigned count = itr->extent, discrepant = 0;
      for ( long pid = itr->pixid; count--; incr(pid,North) ) {
	 if ( _pixEventMap[pid] ) ++discrepant;
	 put( pid, Invalid );
      }

      if ( discrepant ) {
	 errstr << "(CCD " << _ccd << ") "
		<< (uplinked ? "uplinked bad " : "advisory dead ")
		<< (itr->extent > 1 ? "segment " : "pixel ")
		<< "at (" << col(itr->pixid) << "," << row(itr->pixid)
		<< "): ";
	 if ( itr->extent > 1 )
	    errstr << discrepant << " out of " << itr->extent << " pixels ";
	 errstr << "found active in telemetry"
		<< Sas::warning("bogusCCF");
      }
   }
}

inline bool BadPixBuilder::badPixel( unsigned n, unsigned m,
				     float s, unsigned f ) const
{
   return n > m && (n-m)*(n-m) > s*(n*(f-n) + m*(f-m));
}

bool BadPixBuilder::badColumn( const ColumnStats &c0, const ColumnStats &c1,
			       float scale, unsigned frameCount ) const
{
   float diff = c0.average - c1.average;
   if ( diff < 0 ) return false;
   float stat0 = c0.average * (frameCount - c0.average) / c0.pixels;
   float stat1 = c1.average * (frameCount - c1.average) / c1.pixels;
   return diff * diff > scale * (stat0 + stat1);
}

void BadPixBuilder::findHotPixels( unsigned frameCount,
				   float pixsharpness, float colsharpness )
{
   if ( frameCount == 0 ) return;

   float pixFactor = pixsharpness * pixsharpness / frameCount;
   float colFactor = colsharpness * colsharpness / frameCount;
   vector<ColumnStats>::iterator cItr, mItr;
   deque<long> badpix;

   // search both nodes for hot pixels, accumulate column statistics

   for ( unsigned node = 0; node < 2; ++node ) {
      struct Span col = _window[node].col;
      cItr = _columnStats.begin() + col.min - bounds().col.min + 1;
      for (; col.min <= col.max; ++cItr ) {
	 struct Span row = _window[node].row;
	 long nid[4], pid = id( col.min++, row.min );
	 unsigned colEventCount = 0, goodPixels = 0;
	 for (; row.min++ <= row.max; incr(pid,North) ) {
	    unsigned n = _pixEventMap[pid], m = UINT_MAX;

	    if ( n > 0 ) {
	       neighbors(pid,nid);
	       for ( unsigned i = 0; i < 4; ++i )
		  if ( !(*this)[nid[i]] ) minimize( m, _pixEventMap[nid[i]] );
	    }

	    if ( badPixel(n,m,pixFactor,frameCount) ) badpix.push_front(pid);
	    else
	       if ( !(*this)[pid] ) {
		  colEventCount += _colEventMap[pid];
		  ++goodPixels;
	       }
	 }
	 if ( goodPixels ) {
	    cItr->pixels = goodPixels;
	    cItr->average = float(colEventCount) / goodPixels;
	 }
      }
   }

   // mark found hot pixels in bad pixel map

   deque<long>::const_iterator itr;
   for ( itr = badpix.begin(); itr != badpix.end(); ++itr )
      put( *itr, FoundHot );

   // find and mark hot columns

   for ( cItr = _columnStats.begin(); cItr != _columnStats.end(); ++cItr )
      if ( cItr->pixels ) {
	 if ( (cItr+1)->pixels && (cItr-1)->pixels )
	    mItr = cItr + ((cItr+1)->average < (cItr-1)->average ? +1 : -1);
	 else if ( (cItr+1)->pixels )
	    mItr = cItr + 1;
	 else if ( (cItr-1)->pixels )
	    mItr = cItr - 1;
	 else
	    continue;

	 if ( badColumn(*cItr,*mItr,colFactor,frameCount) ) {
	    int col = cItr - _columnStats.begin() - 1 + bounds().col.min;
	    put( Bounds(Span(col),bounds().row), FoundHot );
	 }
      }
}

void BadPixBuilder::markCalBadPix( bool withfoundhot )
{
   vector<CalBadPix>::const_iterator itr;
   for ( itr = _calbp.begin(); itr != _calbp.end(); ++itr ) {
      unsigned int duplicated = 0, discrepant = 0;

      long pid = itr->pixid;
      for ( unsigned count = itr->extent; count--; incr(pid,North) ) {
	 unsigned char tag = (*this)[pid];
	 if ( tag != FoundHot ) {
	    if ( itr->tag == AdvisedHot ) ++discrepant;
	    if ( tag != 0 && tag != Invalid ) ++duplicated;
	 }
	 put( pid, itr->tag );
      }

      if ( withfoundhot && discrepant ) {
	 errstr << "(CCD " << _ccd << ") advisory hot "
		<< (itr->extent > 1 ? "segment" : "pixel")
		<< " at (" << col(itr->pixid) << "," << row(itr->pixid)
		<< "): ";
	 if ( itr->extent > 1 )
	    errstr << discrepant << " out of " << itr->extent << " pixels ";
	 errstr << "not found hot in telemetry"
		<< Sas::warning("discrepancy");
      }

      if ( duplicated ) {
	 errstr << "(CCD " << _ccd << ") CCF bad "
		<< (itr->extent > 1 ? "segment" : "pixel")
		<< " at (" << col(itr->pixid) << "," << row(itr->pixid)
		<< "): ";
	 if ( itr->extent > 1 )
	    errstr << duplicated << " out of " << itr->extent << " pixels ";
	 errstr << "not uniquely defined"
		<< Sas::warning("ccfBadPixDupl");
      }
   }
}

void BadPixBuilder::collect( vector<BadPix> badpix[2] )
{
   unsigned char tag;
   long pid, itr;
   while ( pop(pid,tag) ) {
      unsigned count = 1;
      for ( itr = pid; (*this)[incr(itr,North)] == tag; ++count )
	 (*this)[itr] = 0;
      for ( itr = pid; (*this)[incr(itr,South)] == tag; ++count )
	 (*this)[itr] = 0;

      struct BadPix bp( col(itr), 1+row(itr), count, tag );
      if ( bp.chipX <= _window[0].col.max+1 ) badpix[0].push_back(bp);
      if ( bp.chipX >= _window[1].col.min-1 ) badpix[1].push_back(bp);
   }
}
